# -*- coding: utf-8 -*-

class Proxy(object):
    def __init__(self, target=None):
        super(Proxy, self).__init__()
        self._orig_target = target
        
    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        else:
            target = self._target()
            if target:
                try:
                    return getattr(target, name)
                except AttributeError:
                    raise AttributeError("'%s' object has no attribute '%s'" % (type(self).__name__, name))
            else:
                raise AttributeError('%s object has no attribute "%s", and target is not available.' % (type(self), name))
    
    def _target(self):
        return self._orig_target
    


# -*- coding: utf-8 -*-
from decorated.base.function import Function
from decorated.util import modutil
from threading import RLock
import os

class Synchronized(Function):
    def _init(self, lock):
        self._lock = lock
        
    def _call(self, *args, **kw):
        with self._lock:
            return super(Synchronized, self)._call(*args, **kw)

MemoryLock = RLock

if modutil.module_exists('fcntl'): # fcntl is not supported on google appengine
    import fcntl
    
    class FileLock(object):
        def __init__(self, path):
            self._path = path
            
        def __enter__(self):
            _create_file_if_not_exist(self._path)
            self._fd = os.open(self._path, os.O_RDWR)
            fcntl.flock(self._fd, fcntl.LOCK_EX)
            return self
        
        def __exit__(self, t, v, tb):
            fcntl.flock(self._fd, fcntl.LOCK_UN)
        
def _create_file_if_not_exist(path):
    try:
        f = open(path, 'w')
        f.close()
    except OSError:
        pass
    


# -*- coding: utf-8 -*-
from decorated.base.function import Function
import signal
import time

class TimeoutError(Exception):
    pass

class Timeout(object):
    def __init__(self, seconds):
        self._seconds = seconds
        self._old_handler = None
        self._old_alarm_time = None
        
    def __enter__(self):
        if self._seconds != 0:
            self._old_handler = signal.getsignal(signal.SIGALRM)
            def _timeout(*args):
                raise TimeoutError()
            signal.signal(signal.SIGALRM, _timeout)
            old_alarm = signal.alarm(self._seconds)
            if old_alarm != 0:
                self._old_alarm_time = time.time() + old_alarm
        return self
    
    def __exit__(self, *args):
        signal.alarm(0)
        if self._old_handler is not None:
            signal.signal(signal.SIGALRM, self._old_handler)
        if self._old_alarm_time is not None:
            remain_seconds = int(self._old_alarm_time - time.time())
            if remain_seconds < 1:
                # old alarm is overdue
                # the best we can do is rescheduling it at 1 second later
                remain_seconds = 1
            signal.alarm(remain_seconds)
    
class TimeoutDecorator(Function):
    def _init(self, seconds):
        super(TimeoutDecorator, self)._init()
        self._seconds = seconds
        
    def _call(self, *args, **kw):
        with Timeout(self._seconds):
            return super(TimeoutDecorator, self)._call(*args, **kw)



# -*- coding: utf-8 -*-
from decorated.base.function import Function
from decorated.decorators.remove_extra_args import RemoveExtraArgs
import six

class Conditional(Function):
    def _call(self, *args, **kw):
        if self._test_condition(*args, **kw):
            return super(Conditional, self)._call(*args, **kw)
        
    def _init(self, condition):
        if isinstance(condition, six.string_types):
            self._condition = condition
        elif callable(condition):
            self._condition = RemoveExtraArgs(condition)
        else:
            raise Exception('Condition can only be string or callable.')
        
    def _test_condition(self, *args, **kw):
        d = self._resolve_args(*args, **kw)
        if 'self' in d:
            d['self_'] = d.pop('self')
        if isinstance(self._condition, six.string_types):
            return eval(self._condition, d)
        else:
            return self._condition(**d)
        


# -*- coding: utf-8 -*-
from decorated.base.function import Function, partial
from decorated.decorators.conditional import Conditional
from decorated.decorators.events import Event
from decorated.decorators.instantiate import Instantiate
from decorated.decorators.once import Once, OnceSession
from decorated.decorators.remove_extra_args import RemoveExtraArgs
from decorated.decorators.retries import Retries
from decorated.decorators.synchronized import Synchronized
from decorated.decorators.timeout import TimeoutDecorator

Function = Function
partial = partial

Event = Event
conditional = Conditional
instantiate = Instantiate
once = Once
OnceSession = OnceSession
remove_extra_args = RemoveExtraArgs
retries = Retries
synchronized = Synchronized
timeout = TimeoutDecorator
